from datetime import datetime
from time import sleep

from pycrunch_trace.client.api import trace

from pycrunch_trace.demo.interactive_demo_02 import method_in_another_file
from pycrunch_trace.demo.interactive_demo_03 import show_me_how_to_navigate_using_graph
from pycrunch_trace.demo.interactive_demo_04 import alternative_ways_to_trace


@trace('interactive_demo')
def method_you_want_to_trace(some_number: int, some_string: str):
    # Take a look at the inspector panel.
    # Variable you passed, are recorded there.
    print('Press down arrow on keyboard to step into next line')
    print('You can also use toolbar button [↓] near slider')

    print('Lets now change value of some_number.')
    some_number -= 1
    print('Notice that now, inspector recorded value to be 2.')
    print('Any time you can step to previous line using Up Arrow or [↑] button.')

    print('Now, to the function calls.')
    print('Press Right Arrow on keyboard or [→] button to step into the next method')
    # You can always go back using Left Arrow
    from_external_method = method_in_another_file(some_number)

    print("Press [G], if you haven't open graph yet, ")
    print("All toolbars can be accessed using settings button in the top right corner")

    print("Use [→] to `step into` next method")
    sleep(0.1)
    we_need_to_go_deeper()
    alternative_ways_to_trace()

    print('You can start this tutorial again by dragging slider to the beginning')
    print('  This is similar to scrolling through song in a music player or watching game replay')

    print("That's it. Start tracing now by installing:")
    print("  pip install pycrunch-trace")

    one_last_thing('keyboard shortcuts')

    return str(datetime.utcnow())


def we_need_to_go_deeper():
    print("Sometimes insignificant call time makes it hard to navigate the graph")

    print("Lets take factorial for instance")
    sleep(0.1)

    print("Click at `we_need_to_go_deeper` method in the graph, holding [Command] key")
    # or (Control key on Windows)
    factorial = recursive_factorial(6)

    print("You can scope even deeper in call tree.")
    print("  By clicking with the mouse while holding [Command]")

    print("Entire call stack is drawn under the method you scoped to")

    print("You can exit from the scope same way:")
    print(" By holding [Command] and clicking on outer method")

    # Right arrow [→] to continue
    show_me_how_to_navigate_using_graph()


def recursive_factorial(num: int):
    sleep(0.005)
    if num == 1:
        return num
    else:
        return num * recursive_factorial(num - 1)


def one_last_thing(dummy_variable):
    sleep(0.1)
    print('Most useful keyboard shortcuts:')
    shortcuts = dict()
    shortcuts['[G]'] = 'Toggle Graph View'
    shortcuts['[i]'] = 'Toggle Inspector'

    # Moving forward in debugger:
    regular_debug = dict()
    regular_debug['[→]'] = 'Next Event'
    regular_debug['[↓]'] = 'Step Over; Ignoring method calls'
    regular_debug['[Shift]+[↓]'] = 'Step Out of method'

    # Moving back in debugger:
    time_travel_debug = dict()
    time_travel_debug['[←]'] = 'Previous Event'
    time_travel_debug['[↑]'] = 'Step Over, in backward direction;'
    time_travel_debug['[Shift]+[↑]'] = 'Step Out of method, backward direction, before call'

    mutable_variable = 1
    mutable_variable = example_of_step_out_back(mutable_variable + 1)

    # Just use this:
    #
    #    [↑]
    # [←][↓][→]
    #
    # instead of remembering all the F5...F10
    return mutable_variable


def example_of_step_out_back(mutable_variable: int):
    print(mutable_variable)
    mutable_variable += 1
    print('Now try this: [Shift]+[↑]')
    print('Step Out Back will bring you back, right before current function call.')
    print('  When the `mutable_variable` was equal to 1')
    print('  It is really useful if you need to re-enter method')
    print('  Or check state before function call')
    return mutable_variable + 1


if __name__ == "__main__":
    method_you_want_to_trace(3, 'Welcome To PyTrace')




from time import sleep

from pycrunch_trace.client.api import trace


def alternative_ways_to_trace():
    sleep(0.25)
    print('You can use Trace object to manually start and stop tracing')
    print(' Or by applying @trace decorator to the method')
    print(' See examples bellow')

def example_without_decorators():
    from pycrunch_trace.client.api import Trace

    tracer = Trace()
    tracer.start('recording_name')

    code_under_trace()
    another_code_to_trace()

    tracer.stop()


@trace
def example_with_decorator():
    # Recording will be named the same as the method name
    pass


@trace('this_is_custom_name')
def example_with_custom_name():
    pass


from time import sleep


def method_in_another_file(some_number: int):
    print('File contents are stored as they were at the moment of the recording.')

    print('For easy navigation, press G to open graph panel')
    print('Graph will give overview of method calls and their approximate duration.')

    print('Check this:')
    sleep(0.15)

    print('You can see on the graph that we are at 150ms mark of the recording.')
    print('Current time (since recording start) is also displayed in inspector')
    print('  And on graph panel as a needle')

    print('You can skip next method call by using Down Arrow [↓]')
    print('Or enter into by Right Arrow [→]')
    dummy_method()

    print('Note __return pseudo-variable as we exit from this method')
    return some_number * 3 + 1


def dummy_method():
    print('See how `index` variable is changed as you step through the loop')
    for index in range(5):
        print('If you got trapped in a loop like this, use Shift+Down Arrow to step out of this method')
        print('You can also try Shift+Up Arrow to get to the place where this method was called')
        sleep(0.03)

    # if you are not able to read all text, you can toggle inspector
    #   by pressing [i]



from typing import List

from pycrunch_trace.client.networking.commands import EventsSlice, FileContentSlice
from pycrunch_trace.client.networking.strategies.abstract_strategy import AbstractRecordingStrategy

import pyximport

from pycrunch_trace.events.file_contents_in_protobuf import FileContentsInProtobuf
from pycrunch_trace.file_system import tags

pyximport.install()
from pycrunch_trace.events.native_event_buffer_in_protobuf import NativeEventBufferInProtobuf

from pycrunch_trace.server.incoming_traces import incoming_traces
from pycrunch_trace.server.trace_persistance import TracePersistence


class NativeLocalRecordingStrategy(AbstractRecordingStrategy):
    def __init__(self):
        self.persistence = TracePersistence()
        pass

    def prepare(self):
        # nothing to prepare here
        pass

    def recording_start(self, session_id: str):
        incoming_traces.trace_will_start(session_id)
        self.persistence.initialize_file(session_id)

    def recording_stop(self, session_id: str, files_included: List[str], files_excluded: List[str]):
          #  write json metadata


        self.persistence.recording_complete(session_id, files_included, files_excluded)

    def recording_slice(self, x: EventsSlice):
        incoming_traces.did_receive_more_events(x.session_id, len(x.events))
        bytes_to_disk = NativeEventBufferInProtobuf(x.events, x.files).as_bytes()

        self.persistence.flush_chunk(x.session_id, tags.TRACE_TAG_EVENTS, bytes_to_disk)

    def files_slice(self, x: FileContentSlice):
        bytes_to_disk = FileContentsInProtobuf(x.files).as_bytes()

        self.persistence.update_file_header_files_section(x.session_id, len(bytes_to_disk))
        self.persistence.flush_chunk(x.session_id, tags.TRACE_TAG_FILES, bytes_to_disk)





    def clean(self):
        #  nothing to clean
        pass


cdef class NativeExecutionCursor:
    cdef int file
    cdef int line
    cdef str function_name

cdef class NativeVariables:
    cdef list variables

cdef class NativeStackFrame:
    cdef NativeStackFrame parent
    cdef NativeExecutionCursor cursor
    cdef int id

cdef class NativeCodeEvent:
    cpdef str event_name
    cdef NativeExecutionCursor cursor
    cdef NativeStackFrame stack
    cdef double ts
    cdef NativeVariables locals
    cdef NativeVariables input_variables
    cdef NativeVariables return_variables

cdef class NativeVariable:
    cdef str name
    cdef str value





import time
import collections

from pycrunch_trace.client.networking.commands import EventsSlice, FileContentSlice, StopCommand
from pycrunch_trace.file_system.trace_session import TraceSession
from pycrunch_trace.tracing.file_map import FileMap
from pycrunch_trace.native.native_models cimport NativeCodeEvent, NativeExecutionCursor, NativeVariables, NativeVariable, NativeStackFrame

allowed_types = [int, str, float, dict, type(None), bool]
# allowed_types = [int, str, float,  type(None), bool]


cdef class NativeCallStack:
    cdef object stack
    cdef int last_id
    def __init__(self):
        self.stack = collections.deque()
        self.last_id = 0

    cdef NativeStackFrame enter_frame(self, NativeExecutionCursor execution_cursor):
        cdef NativeStackFrame new_candidate
        parent_frame = self.current_frame()
        # print(f"{execution_cursor.file}:{execution_cursor.line} -> {parent_frame} ")
        new_candidate = NativeStackFrame()
        new_candidate.cursor = execution_cursor
        new_candidate.parent = parent_frame
        new_candidate.id = self.next_id()

        self.stack.append(new_candidate)
        return new_candidate

    cdef int next_id(self):
        self.last_id += 1
        return self.last_id

    cdef NativeStackFrame new_cursor_in_current_frame(self, NativeExecutionCursor new_cursor):
        cdef NativeStackFrame cloned
        cdef NativeStackFrame parent
        parent = self.current_frame()

        cloned = NativeStackFrame()
        cloned.cursor = new_cursor
        if parent is not None:
            cloned.parent = parent.parent
        else:
            cloned.parent = parent

        cloned.id = self.next_id()

        if len(self.stack) > 0:
            self.stack.pop()
            self.stack.append(cloned)
            # self.stack[-1] = stack_frame
        else:
            # session just begin, not yet in any stack
            self.stack.append(cloned)
        return cloned

    cdef exit_frame(self):
        self.stack.pop()

    cdef NativeStackFrame top_level_frame_as_clone(self):
        cdef NativeStackFrame current
        cdef NativeStackFrame new_result
        current = self.current_frame()
        if current:
            new_result = NativeStackFrame()
            new_result.id = self.next_id()
            new_result.parent = current.parent
            new_result.cursor = current.cursor
            return new_result
        else:
            return None
        # ???
        # return current

    cdef NativeStackFrame current_frame(self):
        if len(self.stack) > 0:
            return self.stack[-1]
        return None

cdef class NativeBuffer:
    cdef list _buffer
    def __init__(self):
        self._buffer = []

    cdef void append(self, NativeCodeEvent evt):
        self._buffer.append(evt)

    cdef int count(self):
        return self._buffer.__len__()

    cdef finish_chunk(self):
        clone = self._buffer
        self._buffer = []
        return clone

cdef class NativeTracerPerf:
    cdef int total_samples
    cdef double total_time

    def __init__(self):
        self.total_samples = 1
        self.total_time = 0.00
        pass

    cdef void did_execute_line(self, double ts_diff):
        self.total_samples += 1
        self.total_time += ts_diff

    def print_avg_time(self):
        each = 1
        should_print = self.total_samples % each == 0
        should_print = True
        if should_print:
            time_per_sample = self.total_time / self.total_samples
            print(f'total_samples - {self.total_samples}')
            print(f'total overhead time - {round(self.total_time)} ms')
            print(f'                      {self.total_time}')
            print(f'{time_per_sample:.5f} ms avg call time overhead')

cdef class NativeClock:
    cdef double started_at

    def __init__(self):
        self.started_at = self._system_clock()
        pass

    cdef double now(self):
        cdef double now_without_offset
        now_without_offset = self._system_clock()
        return (now_without_offset - self.started_at) * 1000

    cdef double _system_clock(self):
        return time.perf_counter()
    pass

cdef class NativeTracer:
    cdef int event_number
    cdef int events_so_far
    cdef bint should_trace_variables
    cdef str session_name
    cdef NativeClock clock
    cdef NativeTracerPerf perf
    cdef NativeBuffer event_buffer
    cdef int max_events_before_send
    cdef object file_map
    cdef object file_filter
    cdef object queue
    cdef object session
    cdef NativeCallStack call_stack

    # cpdef ClientQueueThread queue

    def __init__(self, session_name, queue, file_filter):
        self.events_so_far = 0
        self.clock = NativeClock()
        self.perf = NativeTracerPerf()
        self.event_buffer = NativeBuffer()
        self.session_name = session_name
        self.max_events_before_send = 500
        self.file_map = FileMap()
        self.file_filter = file_filter
        self.should_trace_variables = file_filter.should_record_variables()
        self.session = TraceSession()

        self.queue = queue
        self.call_stack = NativeCallStack()

    def simple_tracer(self, frame, str event, arg):
        self.internal_c_call(frame, event, arg)
        return self.simple_tracer

    cdef internal_c_call(self, frame, str evt, arg):
        cdef str func_name
        cdef double entered_at
        cdef double end_at
        cdef double diff
        entered_at = self.clock.now()
        self.wip(entered_at, frame, evt, arg)
        self.events_so_far += 1
        # self.process_events(entered_at, event, frame, arg)
        # self.simulation.save_for_simulator(frame, event, arg)
        # print(f"[{co.co_argcount}]{event}: {func_name} {line_no} -> {arg}")
        # print(f"   {frame.f_locals}")
        end_at = self.clock.now()
        diff = end_at - entered_at
        self.perf.did_execute_line(diff)

    cdef wip(self, double entered_at, frame, str event, arg):
        # cdef boolean will_record_current_event
        cdef NativeCodeEvent current
        cdef NativeVariables input_variables
        cdef NativeVariables return_variables
        cdef NativeVariables locals_v
        cdef NativeExecutionCursor current_cursor
        cdef NativeVariable return_var
        cdef NativeStackFrame stack
        cdef int line_no
        cdef bint will_record_current_event
        cdef str function_name
        file_path_under_cursor = frame.f_code.co_filename

        if not self.file_filter.should_trace(file_path_under_cursor):
            will_record_current_event = False
            self.session.will_skip_file(file_path_under_cursor)
        else:
            will_record_current_event = True
            self.events_so_far += 1
            self.session.did_enter_traceable_file(file_path_under_cursor)

        line_no = frame.f_lineno
        function_name = frame.f_code.co_name
        current = NativeCodeEvent()

        if event == 'call' or event == 'line' or event == 'return':
            file_id = self.file_map.file_id(file_path_under_cursor)
            current_cursor = NativeExecutionCursor()
            current_cursor.file = file_id
            current_cursor.line = line_no
            current_cursor.function_name = function_name

            if event == 'line':
                current.event_name = 'line'
                stack = self.call_stack.new_cursor_in_current_frame(current_cursor)

            if event == 'call':
                current.event_name = 'method_enter'
                stack = self.call_stack.enter_frame(current_cursor)
            if event == 'return':
                current.event_name = 'method_exit'
                self.call_stack.exit_frame()
                stack = self.call_stack.current_frame()
            if not will_record_current_event:
                # do not go any further
                return

            current.cursor = current_cursor
            current.stack = stack
            current.ts = entered_at
            if self.should_trace_variables:
                if event == 'line':
                    locals_v = NativeVariables()
                    locals_v.variables = []
                    self.push_traceable_variables(frame, locals_v)
                    current.locals = locals_v

                if event == 'call':
                    input_variables = NativeVariables()
                    input_variables.variables = []
                    self.push_traceable_variables(frame, input_variables)
                    current.input_variables = input_variables
                if event == 'return':
                    return_variables = NativeVariables()
                    return_variables.variables = []
                    return_var = NativeVariable()
                    return_var.name = '__return'
                    return_var.value = self.ensure_safe_for_serialization(arg)
                    return_variables.variables.append(return_var)
                    current.return_variables = return_variables

            self.add_to_event_buffer(current)

        self.flush_queue_if_full()

    cdef str ensure_safe_for_serialization(self, value):
        current_type = type(value)
        if current_type not in allowed_types:
            return str(current_type)
        # return 'a'
        # todo is this slowdown?
        return str(value)

    cdef push_traceable_variables(self, frame, NativeVariables locals):
        cdef NativeVariable current
        for (name, value) in frame.f_locals.items():
            current = NativeVariable()
            current.name = name
            current.value = self.ensure_safe_for_serialization(value)
            locals.variables.append(current)

    def add_to_event_buffer(self, current):
        # todo: is this caused because of array dynamic size/doubling?
        # if True:
        #     return
        self.event_buffer.append(current)

    # def create_cursor(self, file_path_under_cursor, frame):
    #     file_id = self.file_map.file_id(file_path_under_cursor)
    #     cursor = events.ExecutionCursor(file_id, frame.f_lineno, frame.f_code.co_name)
    #     return cursor

    # def get_execution_stack(self):
    #     return self.call_stack.current_frame()
    #
    # def push_traceable_variables(self, frame, variables):
    #     for (name, value) in frame.f_locals.items():
    #         # todo use variable values diffing
    #         variables.push_variable(name, value)

    def flush_outstanding_events(self):
        self.perf.print_avg_time()
        print(f'total events C: {self.event_buffer.count()}')

        old_buffer = self.event_buffer.finish_chunk()
        # self.perf.print_avg_time()
        #
        self.queue.put_events(EventsSlice(self.session_name, self.event_number, old_buffer, self.file_map.files.copy()))

    def finalize(self):
        print('finalizing native tracer')
        self.queue.put_file_slice(FileContentSlice(self.session_name, self.file_map.files.copy()))

        self.queue.tracing_did_complete(
            self.session_name,
            self.session,
        )

    cdef flush_queue_if_full(self):
        if self.event_buffer.count() >= self.max_events_before_send:
            self.flush_outstanding_events()



from time import sleep


def function_call(**kwargs):
    print(kwargs)
    sleep(0.1)
    sum = kwargs['a'] + kwargs['b']
    return sum


def show_me_how_to_navigate_using_graph():
    print('Lets say we have a function call')
    function_call(a=1, b=2)
    sleep(0.12)

    print('Clicking on function in a graph panel')
    print('  brings you to the beginning of the method')

    print('  Holding Shift and click will navigate you to the END')
    print('    So you can overview return values and state of locals')
    print('    Right before exiting from the method')

    print('You can step out of any function by Shift + Down Arrow')
    return 42




import setuptools
from distutils.core import setup

setup(name='pycrunch-trace',
      version='0.1.6',
      description='PyCrunch Time Travel Debugger',
      url='https://pytrace.com/',
      author='Gleb Sevruk',
      author_email='gleb@pycrunch.com',
      license='MIT',
      keywords=[
          'tracing',
          'debugging',
          'time-travel debugging',
          'record-and-replay debugging',
          'live coding',
      ],
      packages=setuptools.find_packages(),
      setup_requires=['wheel', 'Cython'],
      install_requires=[
          'Cython',
          'jsonpickle',
          'PyYAML',
          'protobuf==3.11.3'
      ],
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Debuggers',
      ],
      project_urls={
          'Source': 'https://github.com/gleb-sevruk/pycrunch-trace/',
          'Funding': 'https://pycrunch.com/donate',
      },
      include_package_data=True,
      zip_safe=False)



from io import FileIO
from typing import Union, Dict

from pycrunch_trace.events.base_event import Event

from pycrunch_trace.native.native_models cimport NativeCodeEvent, NativeVariable, NativeStackFrame

from pycrunch_trace.proto import message_pb2
from pycrunch_trace.tracing.file_map import FileMap

cdef class FileContentsInProtobuf:
    cdef object files

    def __init__(self, files: Dict[str, int]):
        self.files = files

    def as_bytes(self):
        session = message_pb2.FilesInSession()
        for (filename, file_id) in self.files.items():
            current_file_content = message_pb2.FileContent()
            current_file_content.id = file_id
            try:
                with FileIO(filename, 'r') as f:
                    current_file_content.content = f.readall()
            except Exception as e:
                print('error! at store_file_contents: ' + filename)
                print(e)
                current_file_content.content = b'error! at store_file_contents: ' + filename.encode('utf-8')
            session.files.append(current_file_content)

        return session.SerializeToString()
